#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 1
#endif
#ifndef PFXDIST
   #define PFXDIST 0
#endif
#ifndef PFADIST
   #define PFADIST 0
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
/* Need to handle BETA=0 case by assigning y to zero outside of loop */
void ATL_UGEMV(ATL_CINT M, ATL_CINT N, const TYPE *A, ATL_CINT lda,
               const TYPE *X, TYPE *Y)
/*
 * 32x1 unrolled mvn_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N1=(N/1)*1, M32=(M/32)*32, lda1=lda*1;
   ATL_INT j;

   #ifdef BETA0
      for (j=0; j < M; j++)
         Y[j] = ATL_rzero;
   #endif
   for (j=N1; j; j -= 1, A += lda1, X += 1)
   {
      const float *A0=A;
      const register float x0=X[0];
      ATL_INT i;
      prefY(Y+1+1-1);
      for (i=0; i < M32; i += 32)
      {
         const float a0_0=A0[i+0], a1_0=A0[i+1], a2_0=A0[i+2], a3_0=A0[i+3],
                     a4_0=A0[i+4], a5_0=A0[i+5], a6_0=A0[i+6], a7_0=A0[i+7],
                     a8_0=A0[i+8], a9_0=A0[i+9], a10_0=A0[i+10], a11_0=A0[i+11],
                     a12_0=A0[i+12], a13_0=A0[i+13], a14_0=A0[i+14],
                     a15_0=A0[i+15], a16_0=A0[i+16], a17_0=A0[i+17],
                     a18_0=A0[i+18], a19_0=A0[i+19], a20_0=A0[i+20],
                     a21_0=A0[i+21], a22_0=A0[i+22], a23_0=A0[i+23],
                     a24_0=A0[i+24], a25_0=A0[i+25], a26_0=A0[i+26],
                     a27_0=A0[i+27], a28_0=A0[i+28], a29_0=A0[i+29],
                     a30_0=A0[i+30], a31_0=A0[i+31];
         register float y0=Y[i+0], y1=Y[i+1], y2=Y[i+2], y3=Y[i+3], y4=Y[i+4],
                        y5=Y[i+5], y6=Y[i+6], y7=Y[i+7], y8=Y[i+8], y9=Y[i+9],
                        y10=Y[i+10], y11=Y[i+11], y12=Y[i+12], y13=Y[i+13],
                        y14=Y[i+14], y15=Y[i+15], y16=Y[i+16], y17=Y[i+17],
                        y18=Y[i+18], y19=Y[i+19], y20=Y[i+20], y21=Y[i+21],
                        y22=Y[i+22], y23=Y[i+23], y24=Y[i+24], y25=Y[i+25],
                        y26=Y[i+26], y27=Y[i+27], y28=Y[i+28], y29=Y[i+29],
                        y30=Y[i+30], y31=Y[i+31];

         prefX(X);
         y0 += a0_0 * x0;
         prefA(A0);
         y1 += a1_0 * x0;
         y2 += a2_0 * x0;
         y3 += a3_0 * x0;
         y4 += a4_0 * x0;
         y5 += a5_0 * x0;
         y6 += a6_0 * x0;
         y7 += a7_0 * x0;
         y8 += a8_0 * x0;
         y9 += a9_0 * x0;
         y10 += a10_0 * x0;
         y11 += a11_0 * x0;
         y12 += a12_0 * x0;
         y13 += a13_0 * x0;
         y14 += a14_0 * x0;
         y15 += a15_0 * x0;
         y16 += a16_0 * x0;
         prefA(A0);
         y17 += a17_0 * x0;
         y18 += a18_0 * x0;
         y19 += a19_0 * x0;
         y20 += a20_0 * x0;
         y21 += a21_0 * x0;
         y22 += a22_0 * x0;
         y23 += a23_0 * x0;
         y24 += a24_0 * x0;
         y25 += a25_0 * x0;
         y26 += a26_0 * x0;
         y27 += a27_0 * x0;
         y28 += a28_0 * x0;
         y29 += a29_0 * x0;
         y30 += a30_0 * x0;
         y31 += a31_0 * x0;
         Y[i+0] = y0;
         Y[i+1] = y1;
         Y[i+2] = y2;
         Y[i+3] = y3;
         Y[i+4] = y4;
         Y[i+5] = y5;
         Y[i+6] = y6;
         Y[i+7] = y7;
         Y[i+8] = y8;
         Y[i+9] = y9;
         Y[i+10] = y10;
         Y[i+11] = y11;
         Y[i+12] = y12;
         Y[i+13] = y13;
         Y[i+14] = y14;
         Y[i+15] = y15;
         Y[i+16] = y16;
         Y[i+17] = y17;
         Y[i+18] = y18;
         Y[i+19] = y19;
         Y[i+20] = y20;
         Y[i+21] = y21;
         Y[i+22] = y22;
         Y[i+23] = y23;
         Y[i+24] = y24;
         Y[i+25] = y25;
         Y[i+26] = y26;
         Y[i+27] = y27;
         Y[i+28] = y28;
         Y[i+29] = y29;
         Y[i+30] = y30;
         Y[i+31] = y31;
      }
      for (i=M32; i < M; i++)
      {
         const float a0_0=A0[i];
         register float y0 = Y[i];

         y0 += a0_0 * x0;
         Y[i] = y0;
      }
   }
}
