#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 24
#endif
#ifndef PFXDIST
   #define PFXDIST 0
#endif
#ifndef PFADIST
   #define PFADIST 0
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGERK(ATL_CINT M, ATL_CINT N, const double *X, const double *Y, 
               double *A, ATL_CINT lda)
/*
 * 16x24 unrolled r1_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N24=(N/24)*24, M16=(M/16)*16, lda24=lda*24;
   ATL_INT j;

   for (j=N24; j; j -= 24, A += lda24, Y += 24)
   {
      double *A0=A, *A1=A0+lda, *A2=A1+lda, *A3=A2+lda, *A4=A3+lda, *A5=A4+lda,
             *A6=A5+lda, *A7=A6+lda, *A8=A7+lda, *A9=A8+lda, *A10=A9+lda,
             *A11=A10+lda, *A12=A11+lda, *A13=A12+lda, *A14=A13+lda,
             *A15=A14+lda, *A16=A15+lda, *A17=A16+lda, *A18=A17+lda,
             *A19=A18+lda, *A20=A19+lda, *A21=A20+lda, *A22=A21+lda,
             *A23=A22+lda;
      const register double y0=Y[0], y1=Y[1], y2=Y[2], y3=Y[3], y4=Y[4],
                            y5=Y[5], y6=Y[6], y7=Y[7], y8=Y[8], y9=Y[9],
                            y10=Y[10], y11=Y[11], y12=Y[12], y13=Y[13],
                            y14=Y[14], y15=Y[15], y16=Y[16], y17=Y[17],
                            y18=Y[18], y19=Y[19], y20=Y[20], y21=Y[21],
                            y22=Y[22], y23=Y[23];
      ATL_INT i;
      prefY(Y+24+24-1);
      for (i=0; i < M16; i += 16)
      {
         const register double x0=X[i+0], x1=X[i+1], x2=X[i+2], x3=X[i+3],
                               x4=X[i+4], x5=X[i+5], x6=X[i+6], x7=X[i+7],
                               x8=X[i+8], x9=X[i+9], x10=X[i+10], x11=X[i+11],
                               x12=X[i+12], x13=X[i+13], x14=X[i+14],
                               x15=X[i+15];

         prefX(X);
         A0[i+0] += y0 * x0;
         prefA(A0);
         A1[i+0] += y1 * x0;
         prefA(A1);
         A2[i+0] += y2 * x0;
         prefA(A2);
         A3[i+0] += y3 * x0;
         prefA(A3);
         A4[i+0] += y4 * x0;
         prefA(A4);
         A5[i+0] += y5 * x0;
         prefA(A5);
         A6[i+0] += y6 * x0;
         prefA(A6);
         A7[i+0] += y7 * x0;
         prefA(A7);
         A8[i+0] += y8 * x0;
         prefA(A8);
         A9[i+0] += y9 * x0;
         prefA(A9);
         A10[i+0] += y10 * x0;
         prefA(A10);
         A11[i+0] += y11 * x0;
         prefA(A11);
         A12[i+0] += y12 * x0;
         prefA(A12);
         A13[i+0] += y13 * x0;
         prefA(A13);
         A14[i+0] += y14 * x0;
         prefA(A14);
         A15[i+0] += y15 * x0;
         prefA(A15);
         A16[i+0] += y16 * x0;
         prefA(A16);
         A17[i+0] += y17 * x0;
         prefA(A17);
         A18[i+0] += y18 * x0;
         prefA(A18);
         A19[i+0] += y19 * x0;
         prefA(A19);
         A20[i+0] += y20 * x0;
         prefA(A20);
         A21[i+0] += y21 * x0;
         prefA(A21);
         A22[i+0] += y22 * x0;
         prefA(A22);
         A23[i+0] += y23 * x0;
         prefA(A23);
         A0[i+1] += y0 * x1;
         A1[i+1] += y1 * x1;
         A2[i+1] += y2 * x1;
         A3[i+1] += y3 * x1;
         A4[i+1] += y4 * x1;
         A5[i+1] += y5 * x1;
         A6[i+1] += y6 * x1;
         A7[i+1] += y7 * x1;
         A8[i+1] += y8 * x1;
         A9[i+1] += y9 * x1;
         A10[i+1] += y10 * x1;
         A11[i+1] += y11 * x1;
         A12[i+1] += y12 * x1;
         A13[i+1] += y13 * x1;
         A14[i+1] += y14 * x1;
         A15[i+1] += y15 * x1;
         A16[i+1] += y16 * x1;
         A17[i+1] += y17 * x1;
         A18[i+1] += y18 * x1;
         A19[i+1] += y19 * x1;
         A20[i+1] += y20 * x1;
         A21[i+1] += y21 * x1;
         A22[i+1] += y22 * x1;
         A23[i+1] += y23 * x1;
         A0[i+2] += y0 * x2;
         A1[i+2] += y1 * x2;
         A2[i+2] += y2 * x2;
         A3[i+2] += y3 * x2;
         A4[i+2] += y4 * x2;
         A5[i+2] += y5 * x2;
         A6[i+2] += y6 * x2;
         A7[i+2] += y7 * x2;
         A8[i+2] += y8 * x2;
         A9[i+2] += y9 * x2;
         A10[i+2] += y10 * x2;
         A11[i+2] += y11 * x2;
         A12[i+2] += y12 * x2;
         A13[i+2] += y13 * x2;
         A14[i+2] += y14 * x2;
         A15[i+2] += y15 * x2;
         A16[i+2] += y16 * x2;
         A17[i+2] += y17 * x2;
         A18[i+2] += y18 * x2;
         A19[i+2] += y19 * x2;
         A20[i+2] += y20 * x2;
         A21[i+2] += y21 * x2;
         A22[i+2] += y22 * x2;
         A23[i+2] += y23 * x2;
         A0[i+3] += y0 * x3;
         A1[i+3] += y1 * x3;
         A2[i+3] += y2 * x3;
         A3[i+3] += y3 * x3;
         A4[i+3] += y4 * x3;
         A5[i+3] += y5 * x3;
         A6[i+3] += y6 * x3;
         A7[i+3] += y7 * x3;
         A8[i+3] += y8 * x3;
         A9[i+3] += y9 * x3;
         A10[i+3] += y10 * x3;
         A11[i+3] += y11 * x3;
         A12[i+3] += y12 * x3;
         A13[i+3] += y13 * x3;
         A14[i+3] += y14 * x3;
         A15[i+3] += y15 * x3;
         A16[i+3] += y16 * x3;
         A17[i+3] += y17 * x3;
         A18[i+3] += y18 * x3;
         A19[i+3] += y19 * x3;
         A20[i+3] += y20 * x3;
         A21[i+3] += y21 * x3;
         A22[i+3] += y22 * x3;
         A23[i+3] += y23 * x3;
         A0[i+4] += y0 * x4;
         A1[i+4] += y1 * x4;
         A2[i+4] += y2 * x4;
         A3[i+4] += y3 * x4;
         A4[i+4] += y4 * x4;
         A5[i+4] += y5 * x4;
         A6[i+4] += y6 * x4;
         A7[i+4] += y7 * x4;
         A8[i+4] += y8 * x4;
         A9[i+4] += y9 * x4;
         A10[i+4] += y10 * x4;
         A11[i+4] += y11 * x4;
         A12[i+4] += y12 * x4;
         A13[i+4] += y13 * x4;
         A14[i+4] += y14 * x4;
         A15[i+4] += y15 * x4;
         A16[i+4] += y16 * x4;
         A17[i+4] += y17 * x4;
         A18[i+4] += y18 * x4;
         A19[i+4] += y19 * x4;
         A20[i+4] += y20 * x4;
         A21[i+4] += y21 * x4;
         A22[i+4] += y22 * x4;
         A23[i+4] += y23 * x4;
         A0[i+5] += y0 * x5;
         A1[i+5] += y1 * x5;
         A2[i+5] += y2 * x5;
         A3[i+5] += y3 * x5;
         A4[i+5] += y4 * x5;
         A5[i+5] += y5 * x5;
         A6[i+5] += y6 * x5;
         A7[i+5] += y7 * x5;
         A8[i+5] += y8 * x5;
         A9[i+5] += y9 * x5;
         A10[i+5] += y10 * x5;
         A11[i+5] += y11 * x5;
         A12[i+5] += y12 * x5;
         A13[i+5] += y13 * x5;
         A14[i+5] += y14 * x5;
         A15[i+5] += y15 * x5;
         A16[i+5] += y16 * x5;
         A17[i+5] += y17 * x5;
         A18[i+5] += y18 * x5;
         A19[i+5] += y19 * x5;
         A20[i+5] += y20 * x5;
         A21[i+5] += y21 * x5;
         A22[i+5] += y22 * x5;
         A23[i+5] += y23 * x5;
         A0[i+6] += y0 * x6;
         A1[i+6] += y1 * x6;
         A2[i+6] += y2 * x6;
         A3[i+6] += y3 * x6;
         A4[i+6] += y4 * x6;
         A5[i+6] += y5 * x6;
         A6[i+6] += y6 * x6;
         A7[i+6] += y7 * x6;
         A8[i+6] += y8 * x6;
         A9[i+6] += y9 * x6;
         A10[i+6] += y10 * x6;
         A11[i+6] += y11 * x6;
         A12[i+6] += y12 * x6;
         A13[i+6] += y13 * x6;
         A14[i+6] += y14 * x6;
         A15[i+6] += y15 * x6;
         A16[i+6] += y16 * x6;
         A17[i+6] += y17 * x6;
         A18[i+6] += y18 * x6;
         A19[i+6] += y19 * x6;
         A20[i+6] += y20 * x6;
         A21[i+6] += y21 * x6;
         A22[i+6] += y22 * x6;
         A23[i+6] += y23 * x6;
         A0[i+7] += y0 * x7;
         A1[i+7] += y1 * x7;
         A2[i+7] += y2 * x7;
         A3[i+7] += y3 * x7;
         A4[i+7] += y4 * x7;
         A5[i+7] += y5 * x7;
         A6[i+7] += y6 * x7;
         A7[i+7] += y7 * x7;
         A8[i+7] += y8 * x7;
         A9[i+7] += y9 * x7;
         A10[i+7] += y10 * x7;
         A11[i+7] += y11 * x7;
         A12[i+7] += y12 * x7;
         A13[i+7] += y13 * x7;
         A14[i+7] += y14 * x7;
         A15[i+7] += y15 * x7;
         A16[i+7] += y16 * x7;
         A17[i+7] += y17 * x7;
         A18[i+7] += y18 * x7;
         A19[i+7] += y19 * x7;
         A20[i+7] += y20 * x7;
         A21[i+7] += y21 * x7;
         A22[i+7] += y22 * x7;
         A23[i+7] += y23 * x7;
         A0[i+8] += y0 * x8;
         prefA(A0);
         A1[i+8] += y1 * x8;
         prefA(A1);
         A2[i+8] += y2 * x8;
         prefA(A2);
         A3[i+8] += y3 * x8;
         prefA(A3);
         A4[i+8] += y4 * x8;
         prefA(A4);
         A5[i+8] += y5 * x8;
         prefA(A5);
         A6[i+8] += y6 * x8;
         prefA(A6);
         A7[i+8] += y7 * x8;
         prefA(A7);
         A8[i+8] += y8 * x8;
         prefA(A8);
         A9[i+8] += y9 * x8;
         prefA(A9);
         A10[i+8] += y10 * x8;
         prefA(A10);
         A11[i+8] += y11 * x8;
         prefA(A11);
         A12[i+8] += y12 * x8;
         prefA(A12);
         A13[i+8] += y13 * x8;
         prefA(A13);
         A14[i+8] += y14 * x8;
         prefA(A14);
         A15[i+8] += y15 * x8;
         prefA(A15);
         A16[i+8] += y16 * x8;
         prefA(A16);
         A17[i+8] += y17 * x8;
         prefA(A17);
         A18[i+8] += y18 * x8;
         prefA(A18);
         A19[i+8] += y19 * x8;
         prefA(A19);
         A20[i+8] += y20 * x8;
         prefA(A20);
         A21[i+8] += y21 * x8;
         prefA(A21);
         A22[i+8] += y22 * x8;
         prefA(A22);
         A23[i+8] += y23 * x8;
         prefA(A23);
         A0[i+9] += y0 * x9;
         A1[i+9] += y1 * x9;
         A2[i+9] += y2 * x9;
         A3[i+9] += y3 * x9;
         A4[i+9] += y4 * x9;
         A5[i+9] += y5 * x9;
         A6[i+9] += y6 * x9;
         A7[i+9] += y7 * x9;
         A8[i+9] += y8 * x9;
         A9[i+9] += y9 * x9;
         A10[i+9] += y10 * x9;
         A11[i+9] += y11 * x9;
         A12[i+9] += y12 * x9;
         A13[i+9] += y13 * x9;
         A14[i+9] += y14 * x9;
         A15[i+9] += y15 * x9;
         A16[i+9] += y16 * x9;
         A17[i+9] += y17 * x9;
         A18[i+9] += y18 * x9;
         A19[i+9] += y19 * x9;
         A20[i+9] += y20 * x9;
         A21[i+9] += y21 * x9;
         A22[i+9] += y22 * x9;
         A23[i+9] += y23 * x9;
         A0[i+10] += y0 * x10;
         A1[i+10] += y1 * x10;
         A2[i+10] += y2 * x10;
         A3[i+10] += y3 * x10;
         A4[i+10] += y4 * x10;
         A5[i+10] += y5 * x10;
         A6[i+10] += y6 * x10;
         A7[i+10] += y7 * x10;
         A8[i+10] += y8 * x10;
         A9[i+10] += y9 * x10;
         A10[i+10] += y10 * x10;
         A11[i+10] += y11 * x10;
         A12[i+10] += y12 * x10;
         A13[i+10] += y13 * x10;
         A14[i+10] += y14 * x10;
         A15[i+10] += y15 * x10;
         A16[i+10] += y16 * x10;
         A17[i+10] += y17 * x10;
         A18[i+10] += y18 * x10;
         A19[i+10] += y19 * x10;
         A20[i+10] += y20 * x10;
         A21[i+10] += y21 * x10;
         A22[i+10] += y22 * x10;
         A23[i+10] += y23 * x10;
         A0[i+11] += y0 * x11;
         A1[i+11] += y1 * x11;
         A2[i+11] += y2 * x11;
         A3[i+11] += y3 * x11;
         A4[i+11] += y4 * x11;
         A5[i+11] += y5 * x11;
         A6[i+11] += y6 * x11;
         A7[i+11] += y7 * x11;
         A8[i+11] += y8 * x11;
         A9[i+11] += y9 * x11;
         A10[i+11] += y10 * x11;
         A11[i+11] += y11 * x11;
         A12[i+11] += y12 * x11;
         A13[i+11] += y13 * x11;
         A14[i+11] += y14 * x11;
         A15[i+11] += y15 * x11;
         A16[i+11] += y16 * x11;
         A17[i+11] += y17 * x11;
         A18[i+11] += y18 * x11;
         A19[i+11] += y19 * x11;
         A20[i+11] += y20 * x11;
         A21[i+11] += y21 * x11;
         A22[i+11] += y22 * x11;
         A23[i+11] += y23 * x11;
         A0[i+12] += y0 * x12;
         A1[i+12] += y1 * x12;
         A2[i+12] += y2 * x12;
         A3[i+12] += y3 * x12;
         A4[i+12] += y4 * x12;
         A5[i+12] += y5 * x12;
         A6[i+12] += y6 * x12;
         A7[i+12] += y7 * x12;
         A8[i+12] += y8 * x12;
         A9[i+12] += y9 * x12;
         A10[i+12] += y10 * x12;
         A11[i+12] += y11 * x12;
         A12[i+12] += y12 * x12;
         A13[i+12] += y13 * x12;
         A14[i+12] += y14 * x12;
         A15[i+12] += y15 * x12;
         A16[i+12] += y16 * x12;
         A17[i+12] += y17 * x12;
         A18[i+12] += y18 * x12;
         A19[i+12] += y19 * x12;
         A20[i+12] += y20 * x12;
         A21[i+12] += y21 * x12;
         A22[i+12] += y22 * x12;
         A23[i+12] += y23 * x12;
         A0[i+13] += y0 * x13;
         A1[i+13] += y1 * x13;
         A2[i+13] += y2 * x13;
         A3[i+13] += y3 * x13;
         A4[i+13] += y4 * x13;
         A5[i+13] += y5 * x13;
         A6[i+13] += y6 * x13;
         A7[i+13] += y7 * x13;
         A8[i+13] += y8 * x13;
         A9[i+13] += y9 * x13;
         A10[i+13] += y10 * x13;
         A11[i+13] += y11 * x13;
         A12[i+13] += y12 * x13;
         A13[i+13] += y13 * x13;
         A14[i+13] += y14 * x13;
         A15[i+13] += y15 * x13;
         A16[i+13] += y16 * x13;
         A17[i+13] += y17 * x13;
         A18[i+13] += y18 * x13;
         A19[i+13] += y19 * x13;
         A20[i+13] += y20 * x13;
         A21[i+13] += y21 * x13;
         A22[i+13] += y22 * x13;
         A23[i+13] += y23 * x13;
         A0[i+14] += y0 * x14;
         A1[i+14] += y1 * x14;
         A2[i+14] += y2 * x14;
         A3[i+14] += y3 * x14;
         A4[i+14] += y4 * x14;
         A5[i+14] += y5 * x14;
         A6[i+14] += y6 * x14;
         A7[i+14] += y7 * x14;
         A8[i+14] += y8 * x14;
         A9[i+14] += y9 * x14;
         A10[i+14] += y10 * x14;
         A11[i+14] += y11 * x14;
         A12[i+14] += y12 * x14;
         A13[i+14] += y13 * x14;
         A14[i+14] += y14 * x14;
         A15[i+14] += y15 * x14;
         A16[i+14] += y16 * x14;
         A17[i+14] += y17 * x14;
         A18[i+14] += y18 * x14;
         A19[i+14] += y19 * x14;
         A20[i+14] += y20 * x14;
         A21[i+14] += y21 * x14;
         A22[i+14] += y22 * x14;
         A23[i+14] += y23 * x14;
         A0[i+15] += y0 * x15;
         A1[i+15] += y1 * x15;
         A2[i+15] += y2 * x15;
         A3[i+15] += y3 * x15;
         A4[i+15] += y4 * x15;
         A5[i+15] += y5 * x15;
         A6[i+15] += y6 * x15;
         A7[i+15] += y7 * x15;
         A8[i+15] += y8 * x15;
         A9[i+15] += y9 * x15;
         A10[i+15] += y10 * x15;
         A11[i+15] += y11 * x15;
         A12[i+15] += y12 * x15;
         A13[i+15] += y13 * x15;
         A14[i+15] += y14 * x15;
         A15[i+15] += y15 * x15;
         A16[i+15] += y16 * x15;
         A17[i+15] += y17 * x15;
         A18[i+15] += y18 * x15;
         A19[i+15] += y19 * x15;
         A20[i+15] += y20 * x15;
         A21[i+15] += y21 * x15;
         A22[i+15] += y22 * x15;
         A23[i+15] += y23 * x15;
      }
      for (i=M16; i < M; i++)
      {
         const register double x0 = X[i];
         A0[i] += y0 * x0;
         A1[i] += y1 * x0;
         A2[i] += y2 * x0;
         A3[i] += y3 * x0;
         A4[i] += y4 * x0;
         A5[i] += y5 * x0;
         A6[i] += y6 * x0;
         A7[i] += y7 * x0;
         A8[i] += y8 * x0;
         A9[i] += y9 * x0;
         A10[i] += y10 * x0;
         A11[i] += y11 * x0;
         A12[i] += y12 * x0;
         A13[i] += y13 * x0;
         A14[i] += y14 * x0;
         A15[i] += y15 * x0;
         A16[i] += y16 * x0;
         A17[i] += y17 * x0;
         A18[i] += y18 * x0;
         A19[i] += y19 * x0;
         A20[i] += y20 * x0;
         A21[i] += y21 * x0;
         A22[i] += y22 * x0;
         A23[i] += y23 * x0;
      }
   }
/*
 * Do remaining columns with NU=1 cleanup
 */
   for (j=N-N24; j; j--, A += lda, Y++)
   {
      double *A0=A;
      const register double y0=Y[0];
      ATL_INT i;
      prefY(Y+1+1-1);
      for (i=0; i < M16; i += 16)
      {
         const register double x0=X[i+0], x1=X[i+1], x2=X[i+2], x3=X[i+3],
                               x4=X[i+4], x5=X[i+5], x6=X[i+6], x7=X[i+7],
                               x8=X[i+8], x9=X[i+9], x10=X[i+10], x11=X[i+11],
                               x12=X[i+12], x13=X[i+13], x14=X[i+14],
                               x15=X[i+15];

         prefX(X);
         A0[i+0] += y0 * x0;
         prefA(A0);
         A0[i+1] += y0 * x1;
         A0[i+2] += y0 * x2;
         A0[i+3] += y0 * x3;
         A0[i+4] += y0 * x4;
         A0[i+5] += y0 * x5;
         A0[i+6] += y0 * x6;
         A0[i+7] += y0 * x7;
         A0[i+8] += y0 * x8;
         prefA(A0);
         A0[i+9] += y0 * x9;
         A0[i+10] += y0 * x10;
         A0[i+11] += y0 * x11;
         A0[i+12] += y0 * x12;
         A0[i+13] += y0 * x13;
         A0[i+14] += y0 * x14;
         A0[i+15] += y0 * x15;
      }
      for (i=M16; i < M; i++)
      {
         const register double x0 = X[i];
         A0[i] += y0 * x0;
      }
   }
}
