#include "atlas_misc.h"
#ifndef PFYDIST
   #define PFYDIST 1
#endif

#if !defined(ATL_3DNow) && !defined(ATL_SSE1) && \
    (PFADIST != 0 || PFXDIST != 0 || PFYDIST != 0)
   #include "atlas_prefetch.h"
#endif

#if defined(ATL_3DNow) || defined(ATL_SSE1)
   #ifndef PFIY
      #define PFIY prefetchnta
   #endif
   #ifndef PFIX
      #define PFIX prefetchnta
   #endif
   #ifndef PFIA
      #ifdef ATL_3DNow
         #define PFIA prefetchw
      #else
         #define PFIA prefetcht0
      #endif
   #endif
#endif
/*
 * X & A are prefetched in M loop PF[A,X]DIST (in bytes) ahead
 */
#if PFADIST == 0                /* flag for no prefetch */
   #define prefA(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefA(mem) __asm__ __volatile__ \
         (Mstr(PFIA) " %0" : : "m" (*(((char *)(mem))+PFADIST)))
   #else
      #if PFLVL == 2
         #define prefA(mem) ATL_pfl2W(((char*)mem)+PFADIST)
      #else
         #define prefA(mem) ATL_pfl1W(((char*)mem)+PFADIST)
      #endif
   #endif
#endif
#if PFXDIST == 0                /* flag for no prefetch */
   #define prefX(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefX(mem) __asm__ __volatile__ \
         (Mstr(PFIX) " %0" : : "m" (*(((char *)(mem))+PFXDIST)))
   #else
      #if PFLVL == 2
         #define prefX(mem) ATL_pfl2R(((char*)mem)+PFXDIST)
      #else
         #define prefX(mem) ATL_pfl1R(((char*)mem)+PFXDIST)
      #endif
   #endif
#endif
/*
 * Y is prefetched in N-loop, and always fetches next NU piece
 */
#if PFYDIST == 0                /* flag for no prefetch */
   #define prefY(mem)
#else
   #if defined(ATL_3DNow) || defined(ATL_SSE1)
      #define prefY(mem) \
         __asm__ __volatile__ (Mstr(PFIY) " %0" : : "m" (*(((char *)(mem)))))
   #else
      #if PFLVL == 2
         #define prefY(mem) ATL_pfl2R(mem)
      #else
         #define prefY(mem) ATL_pfl1R(mem)
      #endif
   #endif
#endif
#ifndef ATL_CINT
   #define ATL_CINT const int
#endif
#ifndef ATL_INT
   #define ATL_INT int
#endif
void ATL_UGERK(ATL_CINT M0, ATL_CINT N, const double *X, const double *Y, 
               double *A, ATL_CINT lda0)
/*
 * 28x1 unrolled r1_c.
 * Extracted from ATLAS/tune/blas/gemv/atlas-l2g.base
 */
{
   ATL_CINT N1=(N/1)*1, M=M0+M0, M28=((M0/28)*28)<<1,
            lda=lda0+lda0, lda1=lda*1;
   ATL_INT j;

   for (j=N1; j; j -= 1, A += lda1, Y += 2)
   {
      double *A0=A;
      const register double ry0=Y[0], iy0=Y[1];
      ATL_INT i;
      prefY(Y+2+2-1);
      for (i=0; i < M28; i += 56)
      {
         const register double rx0=X[i+0], ix0=X[i+1], rx1=X[i+2], ix1=X[i+3],
                               rx2=X[i+4], ix2=X[i+5], rx3=X[i+6], ix3=X[i+7],
                               rx4=X[i+8], ix4=X[i+9], rx5=X[i+10], ix5=X[i+11],
                               rx6=X[i+12], ix6=X[i+13], rx7=X[i+14],
                               ix7=X[i+15], rx8=X[i+16], ix8=X[i+17],
                               rx9=X[i+18], ix9=X[i+19], rx10=X[i+20],
                               ix10=X[i+21], rx11=X[i+22], ix11=X[i+23],
                               rx12=X[i+24], ix12=X[i+25], rx13=X[i+26],
                               ix13=X[i+27], rx14=X[i+28], ix14=X[i+29],
                               rx15=X[i+30], ix15=X[i+31], rx16=X[i+32],
                               ix16=X[i+33], rx17=X[i+34], ix17=X[i+35],
                               rx18=X[i+36], ix18=X[i+37], rx19=X[i+38],
                               ix19=X[i+39], rx20=X[i+40], ix20=X[i+41],
                               rx21=X[i+42], ix21=X[i+43], rx22=X[i+44],
                               ix22=X[i+45], rx23=X[i+46], ix23=X[i+47],
                               rx24=X[i+48], ix24=X[i+49], rx25=X[i+50],
                               ix25=X[i+51], rx26=X[i+52], ix26=X[i+53],
                               rx27=X[i+54], ix27=X[i+55];

         prefX(X);
         A0[i+0] += ry0*rx0 - iy0*ix0;
         prefA(A0);
         A0[i+0+1] += ry0*ix0 + iy0*rx0;
         A0[i+2] += ry0*rx1 - iy0*ix1;
         A0[i+2+1] += ry0*ix1 + iy0*rx1;
         A0[i+4] += ry0*rx2 - iy0*ix2;
         A0[i+4+1] += ry0*ix2 + iy0*rx2;
         A0[i+6] += ry0*rx3 - iy0*ix3;
         A0[i+6+1] += ry0*ix3 + iy0*rx3;
         A0[i+8] += ry0*rx4 - iy0*ix4;
         prefA(A0);
         A0[i+8+1] += ry0*ix4 + iy0*rx4;
         A0[i+10] += ry0*rx5 - iy0*ix5;
         A0[i+10+1] += ry0*ix5 + iy0*rx5;
         A0[i+12] += ry0*rx6 - iy0*ix6;
         A0[i+12+1] += ry0*ix6 + iy0*rx6;
         A0[i+14] += ry0*rx7 - iy0*ix7;
         A0[i+14+1] += ry0*ix7 + iy0*rx7;
         A0[i+16] += ry0*rx8 - iy0*ix8;
         prefA(A0);
         A0[i+16+1] += ry0*ix8 + iy0*rx8;
         A0[i+18] += ry0*rx9 - iy0*ix9;
         A0[i+18+1] += ry0*ix9 + iy0*rx9;
         A0[i+20] += ry0*rx10 - iy0*ix10;
         A0[i+20+1] += ry0*ix10 + iy0*rx10;
         A0[i+22] += ry0*rx11 - iy0*ix11;
         A0[i+22+1] += ry0*ix11 + iy0*rx11;
         A0[i+24] += ry0*rx12 - iy0*ix12;
         prefA(A0);
         A0[i+24+1] += ry0*ix12 + iy0*rx12;
         A0[i+26] += ry0*rx13 - iy0*ix13;
         A0[i+26+1] += ry0*ix13 + iy0*rx13;
         A0[i+28] += ry0*rx14 - iy0*ix14;
         A0[i+28+1] += ry0*ix14 + iy0*rx14;
         A0[i+30] += ry0*rx15 - iy0*ix15;
         A0[i+30+1] += ry0*ix15 + iy0*rx15;
         A0[i+32] += ry0*rx16 - iy0*ix16;
         prefA(A0);
         A0[i+32+1] += ry0*ix16 + iy0*rx16;
         A0[i+34] += ry0*rx17 - iy0*ix17;
         A0[i+34+1] += ry0*ix17 + iy0*rx17;
         A0[i+36] += ry0*rx18 - iy0*ix18;
         A0[i+36+1] += ry0*ix18 + iy0*rx18;
         A0[i+38] += ry0*rx19 - iy0*ix19;
         A0[i+38+1] += ry0*ix19 + iy0*rx19;
         A0[i+40] += ry0*rx20 - iy0*ix20;
         prefA(A0);
         A0[i+40+1] += ry0*ix20 + iy0*rx20;
         A0[i+42] += ry0*rx21 - iy0*ix21;
         A0[i+42+1] += ry0*ix21 + iy0*rx21;
         A0[i+44] += ry0*rx22 - iy0*ix22;
         A0[i+44+1] += ry0*ix22 + iy0*rx22;
         A0[i+46] += ry0*rx23 - iy0*ix23;
         A0[i+46+1] += ry0*ix23 + iy0*rx23;
         A0[i+48] += ry0*rx24 - iy0*ix24;
         prefA(A0);
         A0[i+48+1] += ry0*ix24 + iy0*rx24;
         A0[i+50] += ry0*rx25 - iy0*ix25;
         A0[i+50+1] += ry0*ix25 + iy0*rx25;
         A0[i+52] += ry0*rx26 - iy0*ix26;
         A0[i+52+1] += ry0*ix26 + iy0*rx26;
         A0[i+54] += ry0*rx27 - iy0*ix27;
         A0[i+54+1] += ry0*ix27 + iy0*rx27;
      }
      for (i=M28; i < M; i += 2)
      {
         const register double rx0 = X[i], ix0 = X[i+1];
         A0[i] += ry0 * rx0 - iy0 * ix0;
         A0[i+1] += ry0 * ix0 + iy0 * rx0;
      }
   }
}
